/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_ALPHABET_ALPHABET_HH__
#define __DEWEYLAB_BIO_ALPHABET_ALPHABET_HH__

#include <string>

namespace deweylab {
namespace bio {
namespace alphabet {

/// An abstract alphabet for a biological sequence.
///
class Alphabet {
public:
	typedef unsigned char encoded_char; ///< A character encoded by an alphabet.
	typedef std::string encoded_string; ///< A string encoded by an alphabet.
	
	virtual ~Alphabet() {}

	/// Returns the number of characters in this alphabet.
	///
	virtual encoded_char getSize() const = 0;

	/// Returns true if the character is in this alphabet.
	///
	virtual bool isMember(const char c) const = 0;

	/// Returns true if all characters in the string are in this alphabet.
	///
	bool isOn(const std::string& s) const;

	/// Returns true if all characters in the iterator range are in
	/// this alphabet.
	///
	template<typename Iterator>
	bool isOn(Iterator begin, Iterator end) const;

	/// Encode a character by a number in [0, getSize() - 1].
	///
	virtual encoded_char encode(const char c) const = 0;

	/// Encode a string.
	///
	virtual encoded_string encode(const std::string& s) const;

	/// Decode an encoded character.
	///
	virtual char decode(const encoded_char c) const = 0;

	/// Decode an encoded string
	///
	virtual std::string decode(const encoded_string& s) const;

	/// Get the full chemical name for a character.
	///
	virtual std::string getName(const char c) const = 0;
};

template<typename Iterator>
bool Alphabet::isOn(Iterator begin, Iterator end) const {
    for (Iterator it = begin; it != end; ++it) {
		if (not isMember(*it)) {
			return false;
		}
    }
    return true;
}

}
}
}

#endif // __DEWEYLAB_BIO_ALPHABET_ALPHABET_HH__
