/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_ALPHABET_ALPHABETCOUNTER_HH__
#define __DEWEYLAB_BIO_ALPHABET_ALPHABETCOUNTER_HH__

#include <vector>
#include "deweylab/bio/alphabet/Alphabet.hh"

namespace deweylab {
namespace bio {
namespace alphabet {

class AlphabetCounter {
public:
	AlphabetCounter(const Alphabet& alphabet);

	// Add characters in sequence S to the counts
	void count(const std::string& s);

	// Add characters in range to the counts
	template<typename Iterator>
	void count(Iterator begin, Iterator end);

	// Add counts from OTHER to this counter
	void addCounts(const AlphabetCounter& other);
	
	// Reset character counts
	void reset();

	// Get the alphabet of this counter
	const Alphabet& getAlphabet() const;
	
	// Get the count for the character RESIDUE
	size_t getCount(char residue) const;

	// Get the count for the encoded character C
	size_t getEncodedCount(char C) const;
	
	// Get the fraction of characters counted that were RESIDUE
	double getFraction(char residue) const;

	// Get the frequency of the encoded character C
	double getEncodedFraction(char c) const;
	
	// Get the total number of characters counted
	size_t getTotal() const;
	
private:
	const Alphabet& alphabet;
	std::vector<size_t> counts;
	size_t total;
};

inline const Alphabet& AlphabetCounter::getAlphabet() const {
	return alphabet;
}

inline size_t AlphabetCounter::getCount(char residue) const {
	return getEncodedCount(alphabet.encode(residue));
}

inline size_t AlphabetCounter::getEncodedCount(char c) const {
	return counts[c];
}

inline size_t AlphabetCounter::getTotal() const { return total; }

inline double AlphabetCounter::getFraction(char residue) const {
	return getEncodedFraction(alphabet.encode(residue));
}

inline double AlphabetCounter::getEncodedFraction(char c) const {
	return getEncodedCount(c) / static_cast<double>(getTotal());
}

template<typename Iterator>
void AlphabetCounter::count(Iterator begin, Iterator end) {
	for (Iterator it = begin; it != end; ++it) {
		++counts.at(alphabet.encode(*it));
	}
	total += std::distance(begin, end);
}

}
}
}

#endif // __DEWEYLAB_BIO_ALPHABET_ALPHABETCOUNTER_HH__
