/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>

#include "deweylab/bio/formats/agp/InputStream.hh"
#include "deweylab/bio/formats/agp/BackwardMapper.hh"

namespace deweylab {
namespace bio {
namespace formats {
namespace agp {

struct AGPIntervalComparer {
    bool operator()(const Record& x,
					const genome::Interval& y) const {
		int chromComp = x.getChrom().compare(y.getChrom());
		return chromComp < 0
			or (chromComp == 0 and x.getEnd() <= y.getStart());
    }
};
	
BackwardMapper::BackwardMapper(std::istream& stream) {
    // Read in AGP file
    InputStream agp_stream(stream);
    Record rec;
    while (agp_stream >> rec) {
		if (not rec.isGap()) {
			agpRecs.push_back(rec);
		}
    }

    // Sort AGP Records by chrom interval
    std::sort(agpRecs.begin(), agpRecs.end());
}
	
void BackwardMapper::map(const genome::Interval& i,
						 std::vector<genome::BasicInterval>& mapped) {
    genome::BasicInterval query(i);
    std::vector<Record>::const_iterator start
		= std::lower_bound(agpRecs.begin(), agpRecs.end(),
						   query, AGPIntervalComparer());
    if (start == agpRecs.end()) {
		return;
    }

    // Find last AGP record that this interval overlaps
    std::vector<Record>::const_iterator end = start;
    while (end != agpRecs.end() and i.overlaps(*end)) {
		++end;
    }			 

    std::vector<Record>::const_iterator it;
    for (it = start; it != end; ++it) {
		genome::BasicInterval untransformedInt(i);

		genome::BasicInterval target(it->getTargetInterval());
		genome::BasicInterval source(it->getSourceInterval());
			
		int startOffset = std::max(genome::Distance(0),
                                   i.getStart() - target.getStart());
		int endOffset = std::min(i.getEnd() - target.getStart(),
								 target.getLength());

		untransformedInt.setChrom(it->getSourceChrom());

		if (it->getSourceStrand() == '+') {
			untransformedInt.setStart(source.getStart() + startOffset);
			untransformedInt.setEnd(source.getStart() + endOffset);
		} else {
			untransformedInt.setStart(source.getEnd() - endOffset);
			untransformedInt.setEnd(source.getEnd() - startOffset);
			untransformedInt.setStrand(untransformedInt.getStrand().opposite());
		}

		mapped.push_back(untransformedInt);
    }
		
}
		
} } } }
