/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>

#include "deweylab/bio/formats/agp/InputStream.hh"
#include "deweylab/bio/formats/agp/ForwardMapper.hh"

namespace deweylab {
namespace bio {
namespace formats {
namespace agp {
	
ForwardMapper::ForwardMapper(std::istream& stream) {
    // Read in AGP file
    InputStream agp_stream(stream);
    Record rec;
    while (agp_stream >> rec) {
		if (not rec.isGap()) {
			contigMap[rec.getSourceChrom()].push_back(rec);
		}
    }
}
	
void ForwardMapper::map(const genome::Interval& i,
						std::vector<genome::BasicInterval>& mapped) {
    AGPMap::iterator it = contigMap.find(i.getChrom());
    if (it == contigMap.end()) {
		return;
    }
		
    AGPList& mappingList = it->second;

	// Linear search over AGP records that have the same source
	// chromosome as the query interval.  This could be improved by
	// using binary search over a sorted vector of records.
    for (AGPList::iterator mapping = mappingList.begin();
		 mapping != mappingList.end(); ++mapping) {
		if (not overlap(i, *mapping)) {
			continue;
		}
			
		genome::BasicInterval j;
			
		j.setChrom(mapping->getChrom());
			
		if (mapping->getSourceStrand() == '+') {
			j.setStart(mapping->getTargetStart() +
					   std::max(i.getStart() - mapping->getSourceStart(),
								genome::Position(0)));
			j.setEnd(mapping->getTargetEnd() +
					 std::min(i.getEnd() - mapping->getSourceEnd(),
							  genome::Position(0)));
			j.setStrand(i.getStrand());
		} else {
			j.setStart(mapping->getTargetStart() +
					   std::max(mapping->getSourceEnd() - i.getEnd(),
								genome::Position(0)));
			j.setEnd(mapping->getTargetEnd() +
					 std::min(mapping->getSourceStart() - i.getStart(),
							  genome::Position(0)));
			j.setStrand(i.getStrand().opposite());
		}
			
		mapped.push_back(j);
    }
}

bool ForwardMapper::overlap(const genome::Interval& i, const Record& rec) {
    return i.getStart() < rec.getSourceEnd()
		and i.getEnd() > rec.getSourceStart()
		and i.getChrom() == rec.getSourceChrom();
}
	
} } } }
