/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "InputStream.hh"
#include "FormatError.hh"

#include "deweylab/util/string.hh"

namespace deweylab {
namespace bio {
namespace formats {
namespace fastq {

void parse_id_line(const std::string& id_line, Record& rec) {
    using util::string::whitespaceChars;
    size_t id_end = id_line.find_first_of(whitespaceChars);
    if (id_end == std::string::npos) {
        rec.id = id_line;
    } else {
        rec.id = id_line.substr(0, id_end);
        size_t description_begin = id_line.find_first_not_of(whitespaceChars, id_end);
        rec.description = id_line.substr(description_begin);
    }
}

InputStream::InputStream(std::istream& stream,
                         const QualityScoreEncoding* encoding) 
    : stream_(stream),
      qualities_encoding_(encoding),
      at_end_(not stream_),
      last_line_() {
}

struct NotInInterval {
    char start, end;
    NotInInterval(char start, char end) : start(start), end(end) {}
    bool operator()(const char c) { return c < start || c > end; }
};

InputStream& InputStream::operator>>(Record& rec) {
    using util::string::toString;
    
    std::string id_line;
    std::string sequence;
    std::string qualities;
    
    // Read title line (if not read already)
    while (stream_ && last_line_.empty()) {
        stream_ >> last_line_;
    }
    if (last_line_.empty()) {
        at_end_ = true;
        return *this;
    }
    if (last_line_[0] != SEQUENCE_TITLE_LINE_PREFIX) {
        throw FormatError("Line does not begin with proper prefix (" +
                          toString(SEQUENCE_TITLE_LINE_PREFIX) + "): " +
                          last_line_);
    }
    id_line = last_line_.substr(1);
    last_line_.clear();

    // Read sequence lines
    stream_ >> sequence;
    if (not stream_) {
        throw FormatError("Error reading sequence for record: " + id_line);
    }
    while (true) {
        stream_ >> last_line_;
        if (not stream_) {
            throw FormatError("Error reading sequence for record: " + id_line);
        }
        if (last_line_[0] == QUALITIES_TITLE_LINE_PREFIX) {
            break;
        }
        sequence += last_line_;
        last_line_.clear();
    }

    // Read qualities id line
    std::string quality_id_line(last_line_.substr(1));
    last_line_.clear();
    if (not quality_id_line.empty() and quality_id_line != id_line) {
        throw FormatError("Quality line id not equal to sequence line id: " 
                          "sequence line id='" + id_line + "' " +
                          "quality line id='" + quality_id_line + "'");
    }

    // Read qualities string
    stream_ >> qualities;
    if (not stream_) {
        throw FormatError("Error reading qualities for record: " + id_line);
    }
    while (true) {
        stream_ >> last_line_;
        
        if (not stream_ or (last_line_[0] == SEQUENCE_TITLE_LINE_PREFIX and
                            qualities.size() >= sequence.size())) {
            break;
        }
        qualities += last_line_;
        last_line_.clear();
    }

    // Check that sequence and quality strings are the same length
    if (sequence.size() != qualities.size()) {
        throw FormatError("Sequence and quality strings are of different length for record: " + id_line);
    }

    // Check that quality scores are within range
    NotInInterval not_in_interval(qualities_encoding_->min_encoded_val(),
                                  qualities_encoding_->max_encoded_val());
    std::string::const_iterator qual_iter = find_if(qualities.begin(),
                                                    qualities.end(),
                                                    not_in_interval);
    if (qual_iter != qualities.end()) {
        throw FormatError("Quality character out of range: " +
                          std::string(qual_iter, qual_iter + 1));
    }
    
    // Construct record
    parse_id_line(id_line, rec);
    rec.sequence = sequence;
    rec.qualities = qualities;
    rec.qualities_encoding = qualities_encoding_;

    return *this;
}

InputStream::operator bool() {
    return not at_end_;
}

bool InputStream::operator!() {
    return at_end_;
}

}
}
}
}
