/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "QualityScoreEncoding.hh"

#include <map>
#include <vector>
#include <cmath>
#include <stdexcept>
#include <limits>

namespace deweylab {
namespace bio {
namespace formats {
namespace fastq {

double round(double x) {
    return (x < 0.0 ? std::floor(x + 0.5) : std::ceil(x - 0.5));
}

SolexaScore phred_to_solexa(PhredScore s) {
    if (s == 0) {
        return std::numeric_limits<SolexaScore>::min();
    } else {
        return s + static_cast<SolexaScore>(round(10.0 * std::log10(1.0 - std::pow(10.0, s / -10.0))));
    }
}

PhredScore solexa_to_phred(SolexaScore s) {
    return s + static_cast<PhredScore>(round(10.0 * std::log10(1.0 + std::pow(10.0, s / -10.0))));
}

typedef std::map<std::string, const QualityScoreEncoding*> NameEncodingMap;
static NameEncodingMap name_encoding_map;
static std::vector<std::string> quality_score_names;

const QualityScoreEncoding*
get_quality_score_encoding(const std::string& name) {
    if (name_encoding_map.empty()) {
        name_encoding_map["sanger"] = &SangerEncoding::instance();
        name_encoding_map["solexa"] = &OldIlluminaEncoding::instance();
        name_encoding_map["illumina"] = &IlluminaEncoding::instance();
    }
    NameEncodingMap::const_iterator it = name_encoding_map.find(name);
    if (it == name_encoding_map.end()) {
        throw std::runtime_error("Invalid quality score encoding name: " +
                                 name);
    } else {
        return it->second;
    }
}

const QualityScoreNameList&
get_quality_score_names() {
    if (quality_score_names.empty()) {
        quality_score_names.push_back("sanger");
        quality_score_names.push_back("solexa");
        quality_score_names.push_back("illumina");
    }
    return quality_score_names;
}

}
}
}
}
