# Copyright (c) 2007
# Colin Dewey (University of Wisconsin-Madison)
# cdewey@biostat.wisc.edu
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
from deweylab.bio.genome import Interval

class FormatError(Exception):
    def __init__(self, problem):
        self.problem = problem
    def __str__(self):
        return problem

class Sequence:
    def __init__(self, src, start, size, strand, srcSize, text):
        self.src = src
        self.start = start
        self.size = size
        self.strand = strand
        self.srcSize = srcSize
        self.text = text

    def getGenome(self):
        return self.src.split('.')[0]

    def getChrom(self):
        srcFields = self.src.split('.', 2)
        return srcFields[len(srcFields) == 2]

    def setChrom(self, chrom):
        srcFields = self.src.split('.', 2)
        srcFields[len(srcFields) == 2] = chrom
        self.src = '.'.join(srcFields)

    def getStart(self):
        if self.strand == '+':
            return self.start
        else:
            return self.srcSize - self.start - self.size

    def setStart(self, start):
        if self.strand == '+':
            self.start = start
        else:
            self.start = self.srcSize - start - self.size

    def getEnd(self):
        return self.getStart() + self.size

    def setEnd(self, end):
        self.setStart(end - self.size)

    def getStrand(self):
        return self.strand

    def setStrand(self, strand):
        absStart = self.getStart()
        self.strand = strand
        self.setStart(absStart)

    def getSourceSize(self):
        return self.srcSize

    def setSourceSize(self, sourceSize):
        absStart = self.getStart()
        self.srcSize = sourceSize
        self.setStart(absStart)

    def getText(self):
        return self.text

    def getInterval(self):
        return Interval(self.getChrom(),
                        self.getStart(),
                        self.getEnd(),
                        self.getStrand())

    def __str__(self):
        return "s %s %d %d %s %d %s" % \
               (self.src, self.start, self.size, self.strand, self.srcSize,
                self.text)

class Alignment:
    def __init__(self, **parameters):
        self.parameters = parameters
        self.sequences = []

    def addSequence(self, s):
        self.sequences.append(s)

    def __maxLength(self, objectList):
        return max(map(len, map(str, objectList)))

    def __str__(self):
        string = ' '.join(['a'] +
                          ["%s=%s" % p for p in self.parameters.items()]) + '\n'
        maxSrcLen = self.__maxLength([s.src for s in self.sequences])
        maxStartLen = self.__maxLength([s.start for s in self.sequences])
        maxSizeLen = self.__maxLength([s.size for s in self.sequences])
        maxSrcSizeLen = self.__maxLength([s.srcSize for s in self.sequences])
        format = "s %%%ds %%%dd %%%dd %%s %%%dd %%s\n" % \
                 (maxSrcLen, maxStartLen, maxSizeLen, maxSrcSizeLen)
        for s in self.sequences:
            string += format % \
                      (s.src, s.start, s.size, s.strand, s.srcSize, s.text)
        return string

class Header:
    def __init__(self, parameters=[], **attributes):
        if "version" not in attributes:
            attributes["version"] = "1"
        if not parameters:
            parameters = [""]
        self.attributes = attributes
        self.parameters = parameters
    def __str__(self):
        headerLine = ' '.join(["##maf"] +
                              ["%s=%s" % p for p in self.attributes.items()])
        parameterLines = map(str, map(Comment, self.parameters))
        return '\n'.join([headerLine] + parameterLines) + '\n'

class Comment:
    def __init__(self, string=""):
        self.string = string
    def __str__(self):
        return "# %s" % self.string

class Iterator:
    """An iterator over a MAF formatted file.

    Provides an iterator over the records of the MAF file
    """

    def __init__(self, handle):
        self.handle = handle

    def __iter__(self):
        return self

    def makeSequence(self, line):
        src, start, size, strand, srcSize, text = line.split()[1:]
        start = long(start)
        size = long(size)
        srcSize = long(srcSize)
        return Sequence(src, start, size, strand, srcSize, text)

    def makeAlignment(self, line):
        params =  dict([field.split('=') for field in line.split()[1:]])
        return Alignment(**params)

    def getNextLine(self, ignoreBlankLines=0):
        while 1:
            line = self.handle.readline()

            # Stop when EOF reached
            if line == "":
                raise StopIteration

            # Skip over comment lines
            if line.startswith('#'):
                continue

            line = line.strip()
            # Skip over blank lines
            if ignoreBlankLines and not line:
                continue

            return line

    def next(self):
        line = self.getNextLine(ignoreBlankLines=1)
        
        if not line.startswith('a'):
            raise FormatError("Line does not start with 'a':\n" + line)

        a = self.makeAlignment(line)

        while 1:
            try:
                line = self.getNextLine()
                if not line.startswith('s'):
                    return a
                a.addSequence(self.makeSequence(line))
            except StopIteration:
                return a
