/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "deweylab/bio/formats/newick/InputStream.hh"

#include "deweylab/util/io.hh"

#include "boost/spirit/include/classic_core.hpp"
#include "boost/spirit/include/classic_lists.hpp"
#include "boost/spirit/include/classic_confix.hpp"
using namespace boost::spirit::classic;

namespace deweylab {
namespace bio {
namespace formats {
namespace newick {

bool isQuoted(const std::string& s) {
	return s.length() > 1 && s[0] == '\'' && s[s.length() - 1] == '\'';
}

std::string unquote(const std::string& s) {
	std::string us(s.begin() + 1, s.end() - 1);
	std::string::iterator insertPos = us.begin();
	std::string::iterator currPos = us.begin();
	while (currPos != us.end()) {
		if (*currPos == '\'') ++currPos;
		*insertPos = *currPos;
		++insertPos;
		++currPos;
	}
	us.resize(insertPos - us.begin());
	return us;
}

std::string underscoresToSpaces(const std::string& s) {
	std::string str = s;
	std::replace(str.begin(), str.end(), '_', ' ');
	return str;
}

typedef std::stack<phylogenetic::TreePtr> TreeStack;

struct PushTreeAction {
	TreeStack& s;
	PushTreeAction(TreeStack& s) : s(s) {}
	void operator()(char) const {
		s.push(phylogenetic::TreePtr(new phylogenetic::Tree()));
	}
	template<class ParamType>
	void operator()(ParamType, ParamType) const {
		s.push(phylogenetic::TreePtr(new phylogenetic::Tree()));
	}
};

struct PopTreeAction {
	TreeStack& s;
	PopTreeAction(TreeStack& s) : s(s) {}
	template<class ParamType>
	void operator()(ParamType, ParamType) const {
		phylogenetic::TreePtr child = s.top();
		s.pop();
		s.top()->addChild(child);
	}
};

struct LabelAction {
	TreeStack& s;
	LabelAction(TreeStack& s) : s(s) {}
	template<class ParamType>
	void operator()(ParamType const& first, ParamType const& last) const {
		std::string str(first, last);
		if (isQuoted(str)) {
			str = unquote(str);
		} else {
			str = underscoresToSpaces(str);
		}
		s.top()->setLabel(str);
	}
};

struct BranchLengthAction {
	TreeStack& s;
	BranchLengthAction(TreeStack& s) : s(s) {}
	void operator()(const double x) const {
		s.top()->setBranchLength(x);
	}
};

struct TreeGrammar : public grammar<TreeGrammar>
{
	TreeStack& stack;
	
	TreeGrammar(TreeStack& stack)
		: stack(stack)
		{}
	
	template <typename ScannerT>
	struct definition {
		definition(TreeGrammar const& self) {
			tree = subtree >> ';';
			
			child_list = '(' >> subtree[PopTreeAction(self.stack)] % ',' >> ')';
			
			subtree =
				eps_p[PushTreeAction(self.stack)]
				>> *comment
				>> ((child_list >> *comment >> !label) | label)
				>> *comment
				>> !(':' >> *comment >> branch_length)
				>> *comment;

			label = (quoted_label | unquoted_label)[LabelAction(self.stack)];
			unquoted_label = lexeme_d[*(graph_p - '(' - ')' - '[' - ']' - '\'' - ':' - ';' - ',')];
			quoted_label = lexeme_d[ch_p('\'') >> *((print_p - '\'') | "''") >> ch_p('\'')];
//			quoted_label = lexeme_d[ch_p('\'') >> *(print_p - '\'') >> ch_p('\'')];
			branch_length = real_p[BranchLengthAction(self.stack)];
			comment = comment_nest_p('[', ']');
		}
		
		rule<ScannerT> tree, child_list, subtree,
			label, unquoted_label, quoted_label,
			comment,
			branch_length;
		rule<ScannerT> const&
		start() const { return tree; }
	};
	
};

void InputStream::init() {
	grammar = new TreeGrammar(stack);
}

InputStream::InputStream(std::istream& stream)
	: s(util::io::readStream(stream)), curr(s.begin()), at_end(false) {
	init();
}

InputStream::InputStream(const std::string& s)
	: s(s), curr(s.begin()), at_end(false) {
	init();
}

InputStream::InputStream(const InputStream& other)
	: s(other.s), curr(other.curr), at_end(other.at_end) {
	init();
}

InputStream::~InputStream() {
	delete grammar;
}

InputStream& InputStream::operator>>(phylogenetic::TreePtr& tree) {
	typedef std::string::const_iterator IteratorT;
	IteratorT last = s.end();
	parse_info<IteratorT> info = parse(curr, last, *grammar, space_p);

	if (info.hit) {
		tree = stack.top();
		stack.pop();
		curr = info.stop;
		++curr;
	} else {
		at_end = true;
	}
	
	return *this;
}

InputStream::operator bool() {
	return not at_end;
}

bool InputStream::operator!() {
	return at_end;
}

}
}
}
}
