/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "deweylab/bio/formats/newick/OutputStream.hh"

#include "deweylab/bio/phylogenetic/TreeVisitor.hh"

namespace deweylab {
namespace bio {
namespace formats {
namespace newick {

bool needsQuotes(const std::string& s) {
	return s.find_first_of("\t\n\r\f\v()[]':;,_") != std::string::npos;
}

std::string quote(const std::string& s) {
	std::string str;
	str.reserve(2 + std::count(s.begin(), s.end(), '\''));
	str += '\'';
	for (std::string::const_iterator it = s.begin(); it != s.end(); ++it) {
		if (*it == '\'') {
			str += '\'';
		}
		str += *it;
	}
	str += '\'';
	return str;
}

std::string spacesToUnderscores(const std::string& s) {
	std::string str = s;
	std::replace(str.begin(), str.end(), ' ', '_');
	return str;
}

void OutputStream::preVisitNode(const ConstTreePtr& t) {
	if (!t->isLeaf()) stream << '(';
}

void OutputStream::postVisitNode(const ConstTreePtr& t) {
	if (!t->isLeaf()) {
		stream << ')';
	}
	if (!t->getLabel().empty()) {
		if (needsQuotes(t->getLabel())) {
			stream << quote(t->getLabel());
		} else {
			stream << spacesToUnderscores(t->getLabel());
		}
	}
	if (write_branch_lengths) {
		stream << ':' << t->getBranchLength();
	}
	add_comma = false;
}

void OutputStream::postVisitTree(const ConstTreePtr& tree) {
	stream << ";\n";
}

void OutputStream::preVisitEdge(const ConstTreePtr& parent,
								const ConstTreePtr& child) {
	if (add_comma) stream << ',';
	add_comma = false;
}

void OutputStream::postVisitEdge(const ConstTreePtr& parent,
								 const ConstTreePtr& child) {
	add_comma = true;
}

OutputStream::OutputStream(std::ostream& stream,
						   bool write_branch_lengths)
	: stream(stream),
	  write_branch_lengths(write_branch_lengths),
	  add_comma(false) {
}

void OutputStream::setWriteBranchLengths(bool b) {
	write_branch_lengths = b;
}

OutputStream& OutputStream::operator<<(ConstTreePtr tree) {
	if (tree != NULL) dfsTraverse(tree);
	return *this;
}

} } } }
