/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "deweylab/bio/formats/sdb/Record.hh"
#include "deweylab/bio/formats/sdb/DB.hh"
#include "deweylab/util/io.hh"
#include "deweylab/util/string.hh"

namespace deweylab {
namespace bio {
namespace formats {
namespace sdb {

Record::Record()
    : db(NULL) {}
	
Record::Record(DB* db)
    : db(db) {}

bool Record::operator==(const Record& other) const {
    return db == other.db
		&& recNum == other.recNum
		&& seqPos == other.seqPos
		&& seqLength == other.seqLength
		&& compressed == other.compressed;
}
	
void Record::write() const {
    if (!(util::io::binary::write(db->strm, title)
		  && util::io::binary::write(db->strm, recNum)
		  && util::io::binary::write(db->strm, leftSize)
		  && util::io::binary::write(db->strm, rightSize)
		  && util::io::binary::write(db->strm, seqLength)
		  && util::io::binary::write(db->strm, seqPos)
		  && util::io::binary::write(db->strm, compressed))) {
		throw std::runtime_error("Error while reading sdb record");
    }
}

void Record::read() {
    if (!(util::io::binary::read(db->strm, title)
		  && util::io::binary::read(db->strm, recNum)
		  && util::io::binary::read(db->strm, leftSize)
		  && util::io::binary::read(db->strm, rightSize)
		  && util::io::binary::read(db->strm, seqLength)
		  && util::io::binary::read(db->strm, seqPos)
		  && util::io::binary::read(db->strm, compressed))) {
		throw std::runtime_error("Error while reading sdb record");
    }
    // Clear sequence cache
    seq.clear();
}
	
uint32_t Record::getBinarySize() const {
    return util::io::binary::size(title)
		+ util::io::binary::size(recNum)
		+ util::io::binary::size(leftSize)
		+ util::io::binary::size(rightSize)
		+ util::io::binary::size(seqLength)
		+ util::io::binary::size(seqPos)
		+ util::io::binary::size(compressed);
}

const std::string& Record::getTitle() const { return title; }
uint32_t Record::getRecNum() const { return recNum; }
uint32_t Record::getLength() const { return seqLength; }

std::string Record::getSeq() {
    return getSeq(0, seqLength);
}

std::string Record::getSeq(const uint32_t start,
						   const uint32_t end,
						   const char strand,
						   const alphabet::Nucleotide& alphabet) {
    if (end > seqLength || end < start) {
		throw OutOfBoundsError("Bad coordinates for " + title
							   + ": " + util::string::toString(start)
							   + "-" + util::string::toString(end));
    }
		
    std::string seq;

    if (start == end) {
		return seq;
    } else if (compressed) {
		uint32_t startByte = start / 2;
		uint32_t offset = start % 2;
		uint32_t endByte = (start == end ? startByte : (end + 1) / 2);
			
		std::string encoded = readSeq(startByte, endByte - startByte);
		seq = alphabet::Nucleotide::nibDecode(encoded).substr(offset, end - start);
    } else {
		seq = readSeq(start, end - start);
    }
		
    if (strand == '-') {
		alphabet.reverseComplementInPlace(seq);
    }
		
    return seq;
}

size_t Record::getCompressedLength() const {
    if (compressed) {
		return (seqLength / 2) + (seqLength % 2);
    } else {
		return seqLength;
    }
}

std::string Record::readSeq(const uint32_t offset,
							const uint32_t length) {
    if (db->cache) {
		if (seq.empty() && seqLength != 0) {
			std::vector<char> buffer(getCompressedLength());
			bufferSeq(0, buffer);
			seq.assign(buffer.begin(), buffer.end());
		}

		if (offset == 0 && length == seq.length()) {
			return seq;
		} else {
			return seq.substr(offset, length);
		}
			
    } else {
		std::vector<char> buffer(length);
		bufferSeq(offset, buffer);
		return std::string(buffer.begin(), buffer.end());
    }
}

void Record::bufferSeq(const uint32_t offset,
					   std::vector<char>& buffer) {
    fseeko(db->strm, seqPos + offset, SEEK_SET);
    if (!util::io::binary::read(db->strm, &buffer[0], buffer.size())) {
		throw std::runtime_error("Error while reading sdb sequence");
    }
}

} } } }
