/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_BIO_PHYLOGENETIC_TREEVISITOR_HH__
#define __DEWEYLAB_BIO_PHYLOGENETIC_TREEVISITOR_HH__

#include "deweylab/bio/phylogenetic/Tree.hh"

#include <queue>

namespace deweylab {
namespace bio {
namespace phylogenetic {

// A visitor for traversing trees in various orders.  Visitors of
// this type are allowed to modify the tree during traversal.
template<typename TreePtrT>
class AbstractTreeVisitor {
public:
	virtual ~AbstractTreeVisitor() {}
	virtual void preVisitTree(const TreePtrT& t) {}
	virtual void postVisitTree(const TreePtrT& t) {}
	virtual void preVisitNode(const TreePtrT& t) {}
	virtual void postVisitNode(const TreePtrT& t) {}
	virtual void preVisitEdge(const TreePtrT& parent, const TreePtrT& child) {}
	virtual void postVisitEdge(const TreePtrT& parent, const TreePtrT& child) {}

    // Traverse a tree in depth-first order
	void dfsTraverse(const TreePtrT& tree);

	// Traverse a tree in breadth-first order
	void bfsTraverse(const TreePtrT& tree);

private:
    // Traverse a tree in depth-first order
	void dfsTraverseNode(const TreePtrT& tree);
};

template<typename TreePtrT>
void AbstractTreeVisitor<TreePtrT>::dfsTraverse(const TreePtrT& tree) {
	preVisitTree(tree);
	dfsTraverseNode(tree);
	postVisitTree(tree);
}

template<typename TreePtrT>
void AbstractTreeVisitor<TreePtrT>::dfsTraverseNode(const TreePtrT& tree) {
	preVisitNode(tree);
	for (ChildList::const_iterator it = tree->getChildren().begin();
		 it != tree->getChildren().end(); ++it) {
		preVisitEdge(tree, *it);
		dfsTraverseNode(*it);
		postVisitEdge(tree, *it);
	}
	postVisitNode(tree);
}

template<typename TreePtrT>
void AbstractTreeVisitor<TreePtrT>::bfsTraverse(const TreePtrT& tree) {
	preVisitTree(tree);
	std::queue<TreePtrT> unprocessed;
	unprocessed.push(tree);
	while (not unprocessed.empty()) {
		TreePtrT next = unprocessed.front();
		preVisitNode(next);
		for (ChildList::const_iterator it = next->getChildren().begin();
			 it != next->getChildren().end(); ++it) {
			unprocessed.push(*it);
			postVisitEdge(next, *it);
		}
		unprocessed.pop();
		postVisitNode(next);
	}
	postVisitTree(tree);
}

typedef AbstractTreeVisitor<TreePtr> TreeVisitor;
typedef AbstractTreeVisitor<ConstTreePtr> ConstTreeVisitor;

} } }
#endif // __DEWEYLAB_BIO_PHYLOGENETIC_TREEVISITOR_HH__
