/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DEWEYLAB_MATH_ALIASMETHODSAMPLER_HH__
#define __DEWEYLAB_MATH_ALIASMETHODSAMPLER_HH__

#include <vector>

namespace deweylab {
namespace math {

template<typename Category, typename RealType = double>
class AliasMethodSampler {
public:
    AliasMethodSampler(const std::vector<double>& probs,
                       const std::vector<Category>& categories);

    template<class UniformRandomNumberGenerator>
    Category operator()(UniformRandomNumberGenerator& urng) const;

private:
    struct Cell {
        RealType cutoff;
        Category category;
        Category alias;

        Cell(RealType cutoff, Category category, Category alias) 
            : cutoff(cutoff), category(category), alias(alias) {
        }
    };

    std::vector<Cell> cells_;
};

template<typename Category, typename RealType = double>
class CdfSampler {
public:
    CdfSampler(const std::vector<double>& probs,
               const std::vector<Category>& categories);

    template<class UniformRandomNumberGenerator>
    Category operator()(UniformRandomNumberGenerator& urng) const;

private:
    struct Cell {
        RealType prob;
        Category category;

        Cell(RealType prob, Category category) 
            : prob(prob), category(category) {
        }

        bool operator<(const Cell& c) const { return prob > c.prob; }
    };

    std::vector<Cell> cells_;
};

template<typename Category = int,
         typename RealType = double,
         typename Strategy = AliasMethodSampler<Category, RealType> >
class categorical_distribution {
public:
    typedef RealType input_type;
    typedef Category result_type;

    explicit categorical_distribution(const std::vector<RealType>& probs,
                                      const std::vector<Category>& categories)
        : strategy_(probs, categories) {
    }
    
    template<class UniformRandomNumberGenerator>
    result_type operator()(UniformRandomNumberGenerator& urng) {
        return strategy_(urng);
    }
    
private:
    Strategy strategy_;
};

template<typename Category, typename RealType>
AliasMethodSampler<Category, RealType>::
AliasMethodSampler(const std::vector<double>& probs,
                   const std::vector<Category>& categories) {
    // Populate cells vector
    const size_t n = categories.size();
    assert(n > 0);
    cells_.reserve(n);
    for (size_t i = 0; i < n; ++i) {
        cells_.push_back(Cell(probs[i] * n, categories[i], categories[i]));
    }

    // Reorder cells vector so that cells with cutoff less than one
    // are at the front
    typedef typename std::vector<Cell>::iterator CellVectorIterator;
    CellVectorIterator it;
    CellVectorIterator gto(cells_.begin());
    for (it = cells_.begin(); it != cells_.end(); ++it) {
        if (it->cutoff < 1.0) {
            if (it != gto) std::swap(*gto, *it);
            ++gto;
        }
    }

    // Compute the cutoffs and aliases
    // gto is already set to the first element with cutoff >= 1.0
    for (it = cells_.begin(); it != cells_.end() && gto != cells_.end(); ++it) {
        it->alias = gto->category;
        gto->cutoff -= (1.0 - it->cutoff);
        if (gto->cutoff < 1.0) ++gto;
    }
}

template<typename Category, typename RealType>
template<class UniformRandomNumberGenerator>
Category AliasMethodSampler<Category, RealType>::
operator()(UniformRandomNumberGenerator& urng) const {
        double u = cells_.size() * urng();
        size_t i = static_cast<size_t>(std::floor(u));
        double p = u - i;
        const Cell& cell = cells_[i];
        return (p < cell.cutoff ? cell.category : cell.alias);
}

template<typename Category, typename RealType>
CdfSampler<Category, RealType>::
CdfSampler(const std::vector<double>& probs,
           const std::vector<Category>& categories) {
    // Populate cells vector
    const size_t n = categories.size();
    assert(n > 0);
    cells_.reserve(n);
    for (size_t i = 0; i < n; ++i) {
        cells_.push_back(Cell(probs[i], categories[i]));
    }

    // Reorder cells so that highest probability cells are at front
    std::sort(cells_.begin(), cells_.end());    
}

template<typename Category, typename RealType>
template<class UniformRandomNumberGenerator>
Category CdfSampler<Category, RealType>::
operator()(UniformRandomNumberGenerator& urng) const {
        double p = urng();
        typedef typename std::vector<Cell>::const_iterator Iter;
        Iter it(cells_.begin());
        while (it != cells_.end() && it->prob < p) {
            p -= it->prob;
            ++it;
        }
        return (it == cells_.end() ? cells_.back().category : it->category);
}

}
}

#endif // __DEWEYLAB_MATH_ALIASMETHODSAMPLER_HH__
