#define BOOST_TEST_MODULE categorical_distribution

#include <boost/test/unit_test.hpp>
#include "boost/random.hpp"
#include "boost/assign/std/vector.hpp"
#include "deweylab/math/categorical_distribution.hh"

BOOST_AUTO_TEST_CASE( test1 )
{
    using namespace boost::assign;
    typedef int Category;
    std::vector<double> probs;
    std::vector<Category> categories;
    probs += 1.0, 0.0, 0.0, 0.0;
    categories += 3, 4, 5, 6;

    typedef boost::mt19937 Rng;
    typedef deweylab::math::categorical_distribution<Category> Dist;
    typedef boost::variate_generator<Rng&, Dist> Generator;

    Rng rng;
    Dist dist(probs, categories);
    Generator gen(rng, dist);

    Category x = gen();
    BOOST_REQUIRE( x == categories.front() );
}

BOOST_AUTO_TEST_CASE( test2 )
{
    typedef int Category;
    std::vector<double> probs;
    std::vector<Category> categories;
    size_t n = 1000;
    for (size_t i = 0; i < n; ++i) {
        probs.push_back(0);
        categories.push_back(i);
    }
    probs.push_back(1);
    categories.push_back(n);

    typedef boost::mt19937 Rng;
    typedef deweylab::math::categorical_distribution<Category> Dist;
    typedef boost::variate_generator<Rng&, Dist> Generator;

    Rng rng;
    Dist dist(probs, categories);
    Generator gen(rng, dist);

    Category x = gen();
    BOOST_REQUIRE( x == categories.back() );
}

BOOST_AUTO_TEST_CASE( test3 )
{
    typedef char Category;
    std::vector<double> probs;
    std::vector<Category> categories;
    size_t n = 6;
    for (size_t i = 0; i < n; ++i) {
        probs.push_back(1.0 / n);
        categories.push_back('A' + i);
    }

    typedef boost::mt19937 Rng;
    typedef deweylab::math::categorical_distribution<Category> Dist;
    typedef boost::variate_generator<Rng&, Dist> Generator;

    Rng rng;
    Dist dist(probs, categories);
    Generator gen(rng, dist);

    for (size_t i = 0; i < 70; ++i) {
        std::cout << gen();
    }
    std::cout << '\n';
}

BOOST_AUTO_TEST_CASE( test4 )
{
    using namespace boost::assign;
    typedef size_t Category;
    std::vector<double> probs;
    std::vector<Category> categories;
    probs += 0.2, 0.2, 0.1, 0.2, 0.1, 0.2;
    categories += 0, 1, 2, 3, 4, 5;

    typedef boost::mt19937 Rng;
    typedef deweylab::math::categorical_distribution<Category> Dist;
    typedef boost::variate_generator<Rng&, Dist> Generator;

    Rng rng;
    Dist dist(probs, categories);
    Generator gen(rng, dist);
    
    std::vector<size_t> counts(categories.size(), 0);
    
    for (size_t i = 0; i < 10000; ++i) {
        ++counts[gen()];
    }

    for (size_t i = 0; i < counts.size(); ++i) {
        std::cout << i << ": " << counts[i] << " ";
    }
    std::cout << '\n';
}

BOOST_AUTO_TEST_CASE( test4_cdf )
{
    using namespace boost::assign;
    typedef size_t Category;
    typedef double RealType;
    std::vector<RealType> probs;
    std::vector<Category> categories;
    probs += 0.2, 0.2, 0.1, 0.2, 0.1, 0.2;
    categories += 0, 1, 2, 3, 4, 5;

    typedef boost::mt19937 Rng;
    typedef deweylab::math::CdfSampler<Category, RealType> Sampler;
    typedef deweylab::math::categorical_distribution<Category, RealType, Sampler> Dist;
    typedef boost::variate_generator<Rng&, Dist> Generator;

    Rng rng;
    Dist dist(probs, categories);
    Generator gen(rng, dist);
    
    std::vector<size_t> counts(categories.size(), 0);
    
    for (size_t i = 0; i < 10000; ++i) {
        ++counts[gen()];
    }

    for (size_t i = 0; i < counts.size(); ++i) {
        std::cout << i << ": " << counts[i] << " ";
    }
    std::cout << '\n';
}

BOOST_AUTO_TEST_CASE( test5 )
{
    using namespace boost::assign;
    typedef size_t Category;
    std::vector<double> probs;
    std::vector<Category> categories;
    probs += 0.2, 0.2, 0.2, 0.2, 0.2;
    categories += 0, 1, 2, 3, 4;

    typedef boost::mt19937 Rng;
    typedef deweylab::math::categorical_distribution<Category> Dist;
    typedef boost::variate_generator<Rng&, Dist> Generator;

    Rng rng;
    Dist dist(probs, categories);
    Generator gen(rng, dist);
    
    std::vector<size_t> counts(categories.size(), 0);
    
    for (size_t i = 0; i < 10000; ++i) {
        ++counts[gen()];
    }

    for (size_t i = 0; i < counts.size(); ++i) {
        std::cout << i << ": " << counts[i] << " ";
    }
    std::cout << '\n';
}
