#define BOOST_TEST_MODULE fastq

#include <iostream>
#include <sstream>
#include <cstdlib>
#include <boost/test/unit_test.hpp>
#include <boost/assign/list_of.hpp>
#include "boost/assign/std/vector.hpp"
#include "deweylab/bio/formats/fastq.hh"
#include "deweylab/filesystem.hh"

#include <boost/foreach.hpp>
#define foreach         BOOST_FOREACH
#define reverse_foreach BOOST_REVERSE_FOREACH

using namespace deweylab::bio::formats::fastq;

std::string test_filename(const std::string& name) {
    return std::string(std::getenv("srcdir")) + "/data/fastq/" + name + ".fastq";
}

void read_fastq_file(const std::string& filename,
                     const QualityScoreEncoding* encoding) {
    using namespace deweylab::filesystem;
    InputFileStream file_stream(filename);
    InputStream fastq_stream(file_stream, encoding);
    Record fastq_record;
    while (fastq_stream >> fastq_record) {
    }
}

void check_error_thrown(const std::string& name) {
    const QualityScoreEncoding* encoding =
        get_quality_score_encoding("sanger");
    BOOST_REQUIRE_THROW(read_fastq_file(test_filename(name), encoding),
						deweylab::bio::formats::fastq::FormatError);
}

BOOST_AUTO_TEST_CASE ( error_diff_ids ) { check_error_thrown("error_diff_ids"); }
BOOST_AUTO_TEST_CASE ( error_double_qual ) { check_error_thrown("error_double_qual"); }
BOOST_AUTO_TEST_CASE ( error_double_seq ) { check_error_thrown("error_double_seq"); }
BOOST_AUTO_TEST_CASE ( error_long_qual ) { check_error_thrown("error_long_qual"); }
BOOST_AUTO_TEST_CASE ( error_no_qual ) { check_error_thrown("error_no_qual"); }
BOOST_AUTO_TEST_CASE ( error_qual_del ) { check_error_thrown("error_qual_del"); }
BOOST_AUTO_TEST_CASE ( error_qual_escape ) { check_error_thrown("error_qual_escape"); }
BOOST_AUTO_TEST_CASE ( error_qual_null ) { check_error_thrown("error_qual_null"); }
BOOST_AUTO_TEST_CASE ( error_qual_space ) { check_error_thrown("error_qual_space"); }
BOOST_AUTO_TEST_CASE ( error_qual_tab ) { check_error_thrown("error_qual_tab"); }
BOOST_AUTO_TEST_CASE ( error_qual_unit_sep ) { check_error_thrown("error_qual_unit_sep"); }
BOOST_AUTO_TEST_CASE ( error_qual_vtab ) { check_error_thrown("error_qual_vtab"); }
BOOST_AUTO_TEST_CASE ( error_short_qual ) { check_error_thrown("error_short_qual"); }
BOOST_AUTO_TEST_CASE ( error_spaces ) { check_error_thrown("error_spaces"); }
BOOST_AUTO_TEST_CASE ( error_tabs ) { check_error_thrown("error_tabs"); }
BOOST_AUTO_TEST_CASE ( error_trunc_at_plus ) { check_error_thrown("error_trunc_at_plus"); }
BOOST_AUTO_TEST_CASE ( error_trunc_at_qual ) { check_error_thrown("error_trunc_at_qual"); }
BOOST_AUTO_TEST_CASE ( error_trunc_at_seq ) { check_error_thrown("error_trunc_at_seq"); }
BOOST_AUTO_TEST_CASE ( error_trunc_in_plus ) { check_error_thrown("error_trunc_in_plus"); }
BOOST_AUTO_TEST_CASE ( error_trunc_in_qual ) { check_error_thrown("error_trunc_in_qual"); }
BOOST_AUTO_TEST_CASE ( error_trunc_in_seq ) { check_error_thrown("error_trunc_in_seq"); }
BOOST_AUTO_TEST_CASE ( error_trunc_in_title ) { check_error_thrown("error_trunc_in_title"); }

void read_fastq_file(const std::string& filename,
                     const QualityScoreEncoding* encoding,
                     std::vector<Record>& recs) {
    using namespace deweylab::filesystem;
    InputFileStream file_stream(filename);
    InputStream fastq_stream(file_stream, encoding);
    Record fastq_record;
    while (fastq_stream >> fastq_record) {
        recs.push_back(fastq_record);
    }
}

void compare_streams(std::istream& stream1,
                     std::istream& stream2) {
    std::string line1, line2;
    while (true) {
        std::getline(stream1, line1);
        std::getline(stream2, line2);
        if (!stream1 || !stream2) break;
        BOOST_CHECK_EQUAL(line1, line2);
    }
    BOOST_CHECK(!stream1);
    BOOST_CHECK(!stream2);
}

void check_reformat(const std::string& test,
                    const std::string& source_format,
                    const std::string& target_format) {

    const std::string source_join = "_original_";
    const std::string target_join = "_as_";

    std::string source_filename(test_filename(test + source_join + source_format));
    std::string target_filename(test_filename(test + target_join + target_format));
    std::vector<Record> recs;
    const QualityScoreEncoding* source_encoding =
        get_quality_score_encoding(source_format);
    const QualityScoreEncoding* target_encoding =
        get_quality_score_encoding(target_format);
    read_fastq_file(source_filename, source_encoding, recs);
    std::ostringstream target_outstream;
    OutputStream fastq_outstream(target_outstream, target_encoding);
    foreach(Record rec, recs) {
        fastq_outstream << rec;
    }
    std::istringstream target_instream(target_outstream.str());
    deweylab::filesystem::InputFileStream reference_instream(target_filename);
    compare_streams(reference_instream, target_instream);
}

BOOST_AUTO_TEST_CASE( reformat_longreads_sanger_as_sanger ) { check_reformat("longreads", "sanger", "sanger"); }
BOOST_AUTO_TEST_CASE( reformat_wrapping_sanger_as_sanger ) { check_reformat("wrapping", "sanger", "sanger"); }
BOOST_AUTO_TEST_CASE( reformat_illumina_full_range_illumina_as_sanger ) { check_reformat("illumina_full_range", "illumina", "sanger"); }
BOOST_AUTO_TEST_CASE( reformat_sanger_full_range_sanger_as_sanger ) { check_reformat("sanger_full_range", "sanger", "sanger"); }
BOOST_AUTO_TEST_CASE( reformat_solexa_full_range_solexa_as_sanger ) { check_reformat("solexa_full_range", "solexa", "sanger"); }
BOOST_AUTO_TEST_CASE( reformat_misc_dna_sanger_as_sanger ) { check_reformat("misc_dna", "sanger", "sanger"); }
BOOST_AUTO_TEST_CASE( reformat_misc_rna_sanger_as_sanger ) { check_reformat("misc_rna", "sanger", "sanger"); }

BOOST_AUTO_TEST_CASE( reformat_longreads_sanger_as_solexa ) { check_reformat("longreads", "sanger", "solexa"); }
BOOST_AUTO_TEST_CASE( reformat_wrapping_sanger_as_solexa ) { check_reformat("wrapping", "sanger", "solexa"); }
BOOST_AUTO_TEST_CASE( reformat_illumina_full_range_illumina_as_solexa ) { check_reformat("illumina_full_range", "illumina", "solexa"); }
BOOST_AUTO_TEST_CASE( reformat_sanger_full_range_sanger_as_solexa ) { check_reformat("sanger_full_range", "sanger", "solexa"); }
BOOST_AUTO_TEST_CASE( reformat_solexa_full_range_solexa_as_solexa ) { check_reformat("solexa_full_range", "solexa", "solexa"); }    
BOOST_AUTO_TEST_CASE( reformat_misc_dna_sanger_as_solexa ) { check_reformat("misc_dna", "sanger", "solexa"); }
BOOST_AUTO_TEST_CASE( reformat_misc_rna_sanger_as_solexa ) { check_reformat("misc_rna", "sanger", "solexa"); }

BOOST_AUTO_TEST_CASE( reformat_longreads_sanger_as_illumina ) { check_reformat("longreads", "sanger", "illumina"); }
BOOST_AUTO_TEST_CASE( reformat_wrapping_sanger_as_illumina ) { check_reformat("wrapping", "sanger", "illumina"); }
BOOST_AUTO_TEST_CASE( reformat_illumina_full_range_illumina_as_illumina ) { check_reformat("illumina_full_range", "illumina", "illumina"); }
BOOST_AUTO_TEST_CASE( reformat_sanger_full_range_sanger_as_illumina ) { check_reformat("sanger_full_range", "sanger", "illumina"); }
BOOST_AUTO_TEST_CASE( reformat_solexa_full_range_solexa_as_illumina ) { check_reformat("solexa_full_range", "solexa", "illumina"); }    
BOOST_AUTO_TEST_CASE( reformat_misc_dna_sanger_as_illumina ) { check_reformat("misc_dna", "sanger", "illumina"); }
BOOST_AUTO_TEST_CASE( reformat_misc_rna_sanger_as_illumina ) { check_reformat("misc_rna", "sanger", "illumina"); }
