/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <stdexcept>
#include <vector>
#include <map>

#include "deweylab/bio/formats/agp/Record.hh"
#include "deweylab/bio/formats/agp/InputStream.hh"
#include "deweylab/util/options.hh"
#include "deweylab/util/string.hh"
using namespace deweylab;
using namespace bio::formats::agp;
using util::string::toString;

class AGPValidationError : public std::runtime_error {
public:
	AGPValidationError(const std::string& what, const Record& rec) :
		std::runtime_error(what + " in record:\n" + toString(rec)) {
	}
};

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	std::string description =
		"Validate AGP input";
	util::options::Parser parser("< agpInput", description);
	parser.parse(argv, argv + argc);

	try {
		typedef std::vector<Record> RecList;
		typedef std::map<std::string, RecList> ChromRecsMap;
		ChromRecsMap chromRecs;
		
		// Read AGP records into map
		InputStream agpStream(std::cin);
		Record rec;
		while (agpStream >> rec) {
			chromRecs[rec.getChrom()].push_back(rec);
		}
		
		ChromRecsMap::iterator it;
		for (it = chromRecs.begin(); it != chromRecs.end(); ++it) {
			const std::string& chrom = it->first;
			RecList& recs = it->second;
			
			// Order the AGP records by contig start coordinate
			std::sort(recs.begin(), recs.end());
			
			bio::genome::Position lastEnd = 0;
			RecList::const_iterator recIt;
			for (recIt = recs.begin(); recIt != recs.end(); ++recIt) {
				const Record& currRec = *recIt;

				// Check that contig coordinates are valid
				if (currRec.getStart() > currRec.getEnd()) {
					throw AGPValidationError("Contig start is greater than "
											 "contig end", currRec);
				}

				// Check that length of source is the same as contig length
				if (currRec.isGap()) {
					if (currRec.gapLength != currRec.getLength()) {
						throw AGPValidationError("Gap length is not equal to "
												 "contig length", currRec);
					}
				} else {
					// Check that source coordinates are valid
					if (currRec.getSourceStart() > currRec.getSourceEnd()) {
						throw AGPValidationError("Source start is greater "
												 "than source end", currRec);
					}
					if (currRec.getSourceInterval().getLength() !=
						currRec.getTargetInterval().getLength()) {
						throw AGPValidationError("Source length is not equal "
												 "to contig length", currRec);
					}
				}

				// Check that the start in the contig is one more than
				// the last end
				if (currRec.getStart() != lastEnd) {
					throw AGPValidationError("Gap or overlap in contig",
											 currRec);
				}

				lastEnd = currRec.getEnd();
			}

			std::cout << chrom << '\t' << lastEnd << '\n';
		}

	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
