/* Copyright (c) 2010
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "deweylab/bio/formats/fastq.hh"
#include "deweylab/util/options.hh"
#include "deweylab/filesystem.hh"

#include <iostream>
#include <algorithm>

using namespace deweylab;
using namespace deweylab::bio::formats;

class Joiner {
public:
    fastq::Record operator()(const fastq::Record& rec1,
                             const fastq::Record& rec2) const {
        std::string id1 = id_with_end_removed(rec1.id);
        std::string id2 = id_with_end_removed(rec2.id);
        if (id1 != id2) {
            throw std::runtime_error("Record IDs do not match:\n" +
                                     rec1.id + "\n" +
                                     rec2.id);
        }
        fastq::Record joined(rec1);
        joined.id = id1;
        joined.sequence += rec2.sequence;
        joined.qualities += rec2.qualities;
        return joined;
    }
    
private:
    static std::string id_with_end_removed(const std::string& id) {
        if (id.size() >= 2 &&
            id[id.size() - 2] == '/' &&
            (id[id.size() - 1] == '1' || id[id.size() - 1] == '2')) {
            return id.substr(0, id.size() - 2);
        } else {
            return id;
        }
    }
};

int main(int argc, const char* argv[]) {
    std::string end1_filename;
    std::string end2_filename;
    std::string encoding_name = "sanger";

    deweylab::util::options::Parser parser("< fastq input");
    parser.setDescription("Join the ends of paired-end reads into "
                          "single long reads");
    parser.addStoreArg("end1_fastq",
                       "FASTQ-formatted file containing the first end reads",
                       end1_filename);
    parser.addStoreArg("end2_fastq",
                       "FASTQ-formatted file containing the second end reads",
                       end2_filename);
    parser.addStoreOpt('e', "encoding",
                       "Quality score encoding scheme",
                       encoding_name);
	parser.parse(argv, argv + argc);
    
	try {
        const fastq::QualityScoreEncoding* encoding =
            fastq::get_quality_score_encoding(encoding_name);

        filesystem::InputFileStream end1_file(end1_filename);
        fastq::InputStream end1_stream(end1_file, encoding);
        fastq::InputStreamIterator end1_begin(end1_stream), end1_end;

        filesystem::InputFileStream end2_file(end2_filename);
        fastq::InputStream end2_stream(end2_file, encoding);
        fastq::InputStreamIterator end2_begin(end2_stream);
        
        fastq::OutputStream joined_stream(std::cout, encoding);
        fastq::OutputStreamIterator joined_begin(joined_stream);

        Joiner joiner;
        std::transform(end1_begin, end1_end, end2_begin, joined_begin, joiner);

    } catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
