/* Copyright (c) 2010
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "deweylab/bio/formats/fastq.hh"
#include "deweylab/util/options.hh"

#include <iostream>

int main(int argc, const char* argv[]) {
    bool output_length = false;
    bool output_description = false;
    bool output_sequence = false;
    bool output_qualities = false;
    bool output_header = false;
    std::string encoding_name = "sanger";

    deweylab::util::options::Parser parser("< fastq input");
    parser.setDescription("Outputs the id of each sequence in the input "
                          "FASTQ file.  Optionally, the description, length, "
                          "sequence, and qualities may be output as "
                          "additional fields.");
    parser.addStoreOpt('e', "input-encoding",
                       "Quality score encoding scheme",
                       encoding_name);
	parser.addStoreTrueOpt('d', "description",
						   "output sequence descriptions",
						   output_description);
	parser.addStoreTrueOpt('l', "length",
						   "output sequence lengths",
						   output_length);
	parser.addStoreTrueOpt('s', "sequence",
						   "output sequences",
						   output_sequence);
	parser.addStoreTrueOpt('q', "qualities",
						   "output qualities",
						   output_qualities);
	parser.addStoreTrueOpt('h', "header",
						   "output header line with field names",
						   output_header);
	parser.parse(argv, argv + argc);
    
	try {
        if (output_header) {
            std::cout << "id";
            if (output_description) std::cout << '\t' << "description";
            if (output_length) std::cout << '\t' << "length";
            if (output_sequence) std::cout << '\t' << "sequence";
            if (output_qualities) std::cout << '\t' << "qualities";
            std::cout << '\n';
        }

        using namespace deweylab::bio::formats::fastq;
        const QualityScoreEncoding* input_encoding =
            get_quality_score_encoding(encoding_name);
        InputStream in_stream(std::cin, input_encoding);
        Record rec;
        while (in_stream >> rec) {
            std::cout << rec.id;
            if (output_description) std::cout << '\t' << rec.description;
            if (output_length) std::cout << '\t' << rec.sequence.length();
            if (output_sequence) std::cout << '\t' << rec.sequence;
            if (output_qualities) std::cout << '\t' << rec.qualities;
            std::cout << '\n';
        }

    } catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
