/* Copyright (c) 2010
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "deweylab/bio/formats/fastq.hh"
#include "deweylab/util/options.hh"
#include "deweylab/filesystem.hh"

#include <iostream>
#include <algorithm>

using namespace deweylab;
using namespace deweylab::bio::formats;

class Splitter {
public:
    Splitter(fastq::OutputStream& end1_stream,
             fastq::OutputStream& end2_stream,
             size_t split_position)
        : end1_stream_(end1_stream),
          end2_stream_(end2_stream),
          split_position_(split_position) {
    }
    
    void operator()(const fastq::Record& rec) {
        size_t rec_split_position;
        if (split_position_ == 0) {
            rec_split_position = rec.sequence.size() / 2;
        } else if (split_position_ <= rec.sequence.size()) {
            rec_split_position = split_position_;
        } else {
            throw std::runtime_error("Record has read length less than "
                                     "split position value: " +
                                     rec.id);
        }
        fastq::Record end1(rec);
        end1.id += "/1";
        end1.sequence = end1.sequence.substr(0, rec_split_position);
        end1.qualities = end1.qualities.substr(0, rec_split_position);

        fastq::Record end2(rec);
        end2.id += "/2";
        end2.sequence = end2.sequence.substr(rec_split_position);
        end2.qualities = end2.qualities.substr(rec_split_position);

        end1_stream_ << end1;
        end2_stream_ << end2;
    }
    
private:
    fastq::OutputStream& end1_stream_;
    fastq::OutputStream& end2_stream_;
    size_t split_position_;
};

int main(int argc, const char* argv[]) {
    std::string end1_filename;
    std::string end2_filename;
    std::string encoding_name = "sanger";
    size_t split_position = 0;

    deweylab::util::options::Parser parser("< fastq input");
    parser.setDescription("Split FASTQ reads into two end reads");
    parser.addStoreArg("end1_output",
                       "File to which to output the first end reads",
                       end1_filename);
    parser.addStoreArg("end2_output",
                       "File to which to output the second end reads",
                       end2_filename);
    parser.addStoreOpt('e', "encoding",
                       "Quality score encoding scheme",
                       encoding_name);
    parser.addStoreOpt('s', "split-position",
                       "Position (length of the first end) at which to split "
                       "the reads.  If set to zero, then reads are split "
                       "in half.",
                       split_position);
	parser.parse(argv, argv + argc);
    
	try {
        const fastq::QualityScoreEncoding* encoding =
            fastq::get_quality_score_encoding(encoding_name);

        fastq::InputStream joined_stream(std::cin, encoding);
        fastq::InputStreamIterator joined_begin(joined_stream), joined_end;

        filesystem::OutputFileStream end1_file(end1_filename);
        fastq::OutputStream end1_stream(end1_file, encoding);

        filesystem::OutputFileStream end2_file(end2_filename);
        fastq::OutputStream end2_stream(end2_file, encoding);

        Splitter splitter(end1_stream, end2_stream, split_position);
        std::for_each(joined_begin, joined_end, splitter);

    } catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
