/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include "deweylab/bio/formats/maf.hh"
#include "deweylab/bio/formats/agp/ForwardMapper.hh"
#include "deweylab/bio/formats/agp/InputStream.hh"
#include "deweylab/bio/genome/Coord.hh"
#include "deweylab/util/options.hh"
#include "deweylab/util/string.hh"   
#include "boost/unordered_map.hpp"
#include "deweylab/filesystem.hh"
using namespace deweylab;
using namespace filesystem;
using namespace bio;
using util::string::toString;
using boost::unordered_map;

typedef boost::unordered_map<std::string, genome::Distance> LengthMap;

void calc_chrom_sizes(const std::string& agp_filename, LengthMap& chrom_sizes) {
	InputFileStream agp_file(agp_filename);
	bio::formats::agp::InputStream agp_stream(agp_file);
	bio::formats::agp::Record rec;
	while (agp_stream >> rec) {
		chrom_sizes[rec.getChrom()] = std::max(rec.getEnd(), 
				                               chrom_sizes[rec.getChrom()]);
	}
}

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	std::string agpFilename;
	
	// Parse command line
	util::options::Parser parser("< mafInput", "Transform MAF input");
	parser.addStoreArg("agpFile", "", agpFilename);
	parser.parse(argv, argv + argc);

	try {
		InputFileStream agpFile(agpFilename);
		bio::formats::agp::ForwardMapper mapper(agpFile);
		
		LengthMap chrom_sizes;
		calc_chrom_sizes(agpFilename, chrom_sizes);

		bio::formats::maf::InputStream inputStream(std::cin);
		bio::formats::maf::OutputStream outputStream(std::cout,
													 inputStream.getHeader());
		
		std::vector<bio::genome::BasicInterval> mapping;
		
		bio::formats::maf::Record rec;
		while (inputStream >> rec) {
			for (size_t i = 0; i < rec.sequences.size(); ++i) {
				bio::formats::maf::Sequence& seq = rec.sequences[i];
				mapping.clear();
				mapper.map(seq, mapping);
				if (mapping.size() != 1) {
					throw std::runtime_error("Could not map interval: " +
											 toString(seq));
				}
				seq.srcSize = chrom_sizes[mapping.front().getChrom()];
				seq.setInterval(mapping.front());
			}

			outputStream << rec;
		}

	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
