/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	 02111-1307	 USA
*/

#include <iostream>
#include <list>

#include "deweylab/bio/phylogenetic/Tree.hh"
#include "deweylab/bio/phylogenetic/TreeVisitor.hh"
#include "deweylab/bio/formats/newick.hh"
#include "deweylab/util/options.hh"
#include "deweylab/util/string.hh"
using namespace deweylab;

using namespace bio;
using namespace formats;

class LengthPrinter : public phylogenetic::ConstTreeVisitor {
private:
	bool sum;
	std::string sep;
	std::list<std::string> lengths;
	double total;
public:
	LengthPrinter(bool sum, const std::string& sep) : sum(sum), sep(sep) {}

	void preVisitNode(const phylogenetic::ConstTreePtr& t) {
		lengths.push_back(util::string::toString(t->getBranchLength()));
		total += t->getBranchLength();
	}

	std::string operator()(const phylogenetic::TreePtr& t) {
		lengths.clear();
		total = 0;
		dfsTraverse(t);
		if (sum) {
			return util::string::toString(total);
		} else {
			return util::string::join(lengths.begin(), lengths.end(), sep);
		}
	}
};

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	bool sum = false;
	std::string sep = " ";

	// Parse command line
	util::options::Parser parser("< newickInput",
								 "Output the branch lengths in the tree");
	parser.addStoreTrueOpt('t', "total",
						   "Only output sum of branch lengths in tree",
						   sum);
	parser.addStoreOpt('s', "sep",
					   "String separator between branch lengths",
					   sep, "STRING");
	parser.parse(argv, argv + argc);

	try {
		newick::InputStream in_stream(std::cin);
		newick::InputStreamIterator in_begin(in_stream), in_end;
		std::ostream_iterator<std::string> out(std::cout, "\n");
		
		std::transform(in_begin, in_end, out, LengthPrinter(sum, sep));

	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}
	
	return EXIT_SUCCESS;
}
