/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include "deweylab/bio/formats/newick.hh"
#include "deweylab/bio/phylogenetic/TreeVisitor.hh"
#include "deweylab/util/options.hh"
using namespace deweylab;
using namespace bio::formats;
using namespace bio::phylogenetic;

struct LabelRemover : public TreeVisitor {
	void preVisitNode(const TreePtr& t) { t->setLabel(""); }
};

struct InternalLabelRemover : public TreeVisitor {
	void preVisitNode(const TreePtr& t) {
		if (not t->isLeaf()) t->setLabel("");
	}
};

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	bool no_branch_lengths = false;
	bool no_internal_labels = false;
	bool no_labels = false;

	// Parse command line
	util::options::Parser parser("< newickInput > newickOutput",
								 "Reformat newick formatted trees");
	parser.addStoreTrueOpt(0, "no-lengths",
						   "do not output branch lengths",
						   no_branch_lengths);
	parser.addStoreTrueOpt(0, "no-labels",
						   "do not output labels",
						   no_labels);
	parser.addStoreTrueOpt(0, "no-int-labels",
						   "do not output internal node labels",
						   no_internal_labels);
	parser.parse(argv, argv + argc);

	try {
		newick::InputStream in_stream(std::cin);
		newick::OutputStream out_stream(std::cout, not no_branch_lengths);

		TreePtr tree;
		while (in_stream >> tree) {
			if (no_labels) {
				LabelRemover label_visitor;
				label_visitor.dfsTraverse(tree);
			} else if (no_internal_labels) {
				InternalLabelRemover int_label_visitor;
				int_label_visitor.dfsTraverse(tree);
			}
			out_stream << tree;
		}
		
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}
	
	return EXIT_SUCCESS;
}
