/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include "boost/unordered_set.hpp"

#include "deweylab/bio/phylogenetic/Tree.hh"
#include "deweylab/bio/formats/newick.hh"
#include "deweylab/util/options.hh"
using namespace deweylab;

using namespace bio;
using namespace formats;
using boost::unordered_set;

std::string underscoresToSpaces(const std::string& s) {
	std::string str = s;
	std::replace(str.begin(), str.end(), '_', ' ');
	return str;
}

struct SubtreeExtractor {
	boost::unordered_set<std::string> taxaSet;
	bool keepDescendants;

	SubtreeExtractor(std::vector<std::string> taxa,
					 bool keepDescendants)
		: keepDescendants(keepDescendants) {
		for (size_t i = 0; i < taxa.size(); ++i) {
			taxaSet.insert(underscoresToSpaces(taxa[i]));
		}
	}

	phylogenetic::TreePtr operator()(phylogenetic::ConstTreePtr tree) {
		return tree->getSubtree(taxaSet, keepDescendants);
	}
};
	
int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	bool keepDescendants = false;
	std::vector<std::string> taxa;

	// Parse command line
	util::options::Parser parser("< newickInput",
								 "Output tree with given taxa as "
								 "contained in newick input");
	parser.addStoreTrueOpt('d', "keep-descendants", 
						   "include in the output tree all descendants of "
						   "given taxa", keepDescendants);
	parser.addAppendArg("taxon", "Taxon to include in the output tree",
						taxa);
	parser.parse(argv, argv + argc);

	try {
		newick::InputStream in_stream(std::cin);
		newick::InputStreamIterator in_begin(in_stream), in_end;
		newick::OutputStream out_stream(std::cout);
		newick::OutputStreamIterator out_begin(out_stream);
		
		SubtreeExtractor extractor(taxa, keepDescendants);

		std::transform(in_begin, in_end, out_begin, extractor);
		
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}
	
	return EXIT_SUCCESS;
}
