/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <list>

#include "deweylab/bio/phylogenetic/Tree.hh"
#include "deweylab/bio/phylogenetic/TreeVisitor.hh"
#include "deweylab/bio/formats/newick.hh"
#include "deweylab/util/options.hh"
#include "deweylab/util/string.hh"
using namespace deweylab;

using namespace bio;
using namespace formats;

class TaxaPrinter : public phylogenetic::ConstTreeVisitor {
private:
	bool all;
	std::string sep;
	std::list<std::string> labels;
public:
	TaxaPrinter(bool all, const std::string& sep) : all(all), sep(sep) {}

	void preVisitNode(const phylogenetic::ConstTreePtr& t) {
		if ((all or t->isLeaf()) and not t->getLabel().empty()) {
			labels.push_back(t->getLabel());
		}
	}

	std::string operator()(const phylogenetic::TreePtr& t) {
		labels.clear();
		dfsTraverse(t);
		return util::string::join(labels.begin(), labels.end(), sep);
	}
};

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	bool all = false;
	std::string sep = " ";

	// Parse command line
	util::options::Parser parser("< newickInput",
								 "Output the names of the taxa in the tree");
	parser.addStoreTrueOpt('a', "all",
						   "Print all node labels, not just leaf labels", all);
	parser.addStoreOpt('s', "sep",
					   "String separator between taxa names",
					   sep, "STRING");
	parser.parse(argv, argv + argc);

	try {
		newick::InputStream in_stream(std::cin);
		newick::InputStreamIterator in_begin(in_stream), in_end;
		std::ostream_iterator<std::string> out(std::cout, "\n");
		
		std::transform(in_begin, in_end, out, TaxaPrinter(all, sep));

	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}
	
	return EXIT_SUCCESS;
}
