/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string>
#include <iostream>

#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/bio/formats/agp.hh"
#include "deweylab/util/string.hh"
#include "deweylab/util/options.hh"
#include "deweylab/filesystem.hh"

using namespace deweylab;
using namespace filesystem;
using namespace bio::formats;

struct Chunker {
    size_t __max_size;
    fasta::OutputStream __fasta_stream;
    agp::OutputStream __agp_stream;
    bool __output_agp;
    
    Chunker(size_t max_size,
            std::ostream& fasta_stream,
            std::ofstream& agp_stream) 
        : __max_size(max_size),
          __fasta_stream(fasta_stream),
          __agp_stream(agp_stream),
          __output_agp(agp_stream.is_open())
        {}

    void operator()(const fasta::Record& rec);
};

void Chunker::operator()(const fasta::Record& rec) {
    for (size_t i = 0; i * __max_size < rec.sequence.size(); ++i) {
        size_t start = i * __max_size;
        std::string id = rec.id + "." + util::string::toString(i);
        std::string seq = rec.sequence.substr(start, __max_size);
        fasta::Record chunked(id, rec.description, seq);
        __fasta_stream << chunked;
        if (__output_agp) {
            __agp_stream << agp::Record(rec.id, start + 1,
                                        start + seq.size(),
                                        i + 1, agp::Constants::OTHER,
                                        id, 1, seq.size(), '+');
        }
    }
}

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

    // Command-line options
    size_t max_size = 1000000;
    std::string agp_filename;

	// Parse command line
	util::options::Parser parser("< fastaInput",
                                 "Breaks FASTA-formatted sequences into "
                                 "shorter sequences such that no sequence "
                                 "is longer than the specified maximum size.");
	parser.addStoreOpt('s', "max-size",
					   "Maximum length of a sequence.  Sequences longer than "
                       "S will be chunked into multiple sequences.",
					   max_size, "S");
	parser.addStoreOpt('a', "agp",
					   "Name of file to write in AGP format describing how to "
                       "reconstruct original sequences from chunked sequences.",
					   agp_filename, "FILENAME");
	parser.parse(argv, argv + argc);

	try {
        fasta::InputStream in_stream(std::cin);
        fasta::InputStreamIterator in_begin(in_stream), in_end;
        OutputFileStream agp_file;
        if (not agp_filename.empty()) agp_file.open(agp_filename);
        Chunker chunker(max_size, std::cout, agp_file);
        std::for_each(in_begin, in_end, chunker);
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
