/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include "boost/shared_ptr.hpp"

#include "deweylab/app/FastaProcessor.hh"
#include "deweylab/app/FastStandardIOApp.hh"
#include "deweylab/util/string.hh"
#include "deweylab/filesystem.hh"
using namespace deweylab;
using namespace app;
using namespace bio::formats;

class MaskMerger : public FastaWriter, public FastStandardIOApp {
public:
    MaskMerger();
    void run();
protected:
    void merge_mask(std::string& merged, const std::string& masked_seq) const;
    typedef std::vector<std::string> FileList;
    typedef std::vector<fasta::InputStream> StreamList;
    FileList _filenames;
    char _masking_char;
};

void MaskMerger::merge_mask(std::string& merged,
                            const std::string& masked_seq) const {
    for (size_t i = 0; i < merged.length(); ++i) {
        if (masked_seq[i] == _masking_char) { merged[i] = _masking_char; }
    }
}

MaskMerger::MaskMerger() {
    parser.setStandardOutputName("fasta_output");
    parser.setDescription("Merge multiple hardmaskings of the same set of "
                          "sequences.  Sequences must be in the same order in "
                          "all files.");
    parser.addStoreOpt(
        'c', "masking-char",
        "Character used for hardmasked sequence positions",
        _masking_char = 'N');
    parser.addAppendArg(
        "hardmasked_fasta_file",
        "FASTA file with hardmasked sequences",
        _filenames, 1);
}

void MaskMerger::run() {
    // Open fasta streams
    typedef std::vector<filesystem::InputFileStream> FileStreamList;
    typedef boost::shared_ptr<fasta::InputStream> FastaStreamPtr;
    typedef std::vector<FastaStreamPtr> FastaStreamList;
    FileStreamList file_streams(_filenames.begin(), _filenames.end());
    FastaStreamList fasta_streams;
    for (size_t i = 0; i < file_streams.size(); ++i) {
        FastaStreamPtr stream_ptr(new fasta::InputStream(file_streams[i]));
        fasta_streams.push_back(stream_ptr);
    }

    fasta::Record rec;
    while ((*fasta_streams.front()) >> rec) {
        // Use record from first file to start
        fasta::Record merged(rec);
        for (size_t i = 1; i < fasta_streams.size(); ++i) {
            (*fasta_streams[i]) >> rec;

            // Check for error conditions
            if (not (*fasta_streams[i])) {
                throw std::runtime_error(_filenames[i] + " has fewer records");
            } else if (rec.id != merged.id) {
                throw std::runtime_error("Expected record '" + merged.id +
                                         "' but found record '" + rec.id +
                                         "' in " + _filenames[i]);
            } else if (rec.sequence.length() != merged.sequence.length()) {
                throw std::runtime_error("Record '" + merged.id +
                                         "' has a different length in "
                                         + _filenames[i]);
            }

            merge_mask(merged.sequence, rec.sequence);
        }
        write_record(merged);
    }
}

int main(int argc, const char* argv[]) {
    MaskMerger app;
    return app.main(argc, argv);
}
