/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/filesystem.hh"
#include "boost/unordered_map.hpp"
#include "deweylab/util/options.hh"
using namespace deweylab;
using namespace filesystem;
using namespace bio::formats;

struct LengthError {};

std::string dna_align_from_prot_align(const std::string& prot_align,
									  const std::string& dna) {
	const size_t CODON_LEN = 3;
	const char SPACE = '-';
	
	std::string dna_align;
	size_t prot_pos = 0;
	size_t dna_pos = 0;
	for (size_t i = 0; i < prot_align.size(); ++i) {
		if (prot_align[i] == SPACE) {
			dna_align.append(CODON_LEN, SPACE);
		} else {
			if (dna_pos > dna.size()) {
				throw LengthError();
			}

			size_t remaining_len = dna.size() - dna_pos;
			if (remaining_len < CODON_LEN) {
				dna_align.append(dna, dna_pos, remaining_len);
				dna_align.append(CODON_LEN - remaining_len, SPACE);
			} else {
				dna_align.append(dna, dna_pos, CODON_LEN);
			}

			dna_pos += CODON_LEN;
			++prot_pos;
		}
	}
	return dna_align;
}

struct RecordNotFoundError : public std::runtime_error {
	RecordNotFoundError(const std::string& title) 
		: std::runtime_error("DNA record not found: " + title) {}
};

class AlignmentConverter {
public:
	AlignmentConverter(const std::string& dna_filename) {
		InputFileStream dna_file(dna_filename);
		fasta::InputStream fasta_stream(dna_file);
		fasta::Record rec;
		while (fasta_stream >> rec) {
			dna_map[rec.id] = rec.sequence;
		}
	}

	fasta::Record operator()(const fasta::Record& prot_rec) {
		if (dna_map.find(prot_rec.id) == dna_map.end()) {
			throw RecordNotFoundError(prot_rec.id);
		}
		try {
			return fasta::Record(prot_rec.id,
                                 prot_rec.description,
								 dna_align_from_prot_align(prot_rec.sequence,
														   dna_map[prot_rec.id]));
		} catch (const LengthError& e) {
			throw std::runtime_error("DNA sequence not long enough: " +
									 prot_rec.id);
		}
	}

private:
	boost::unordered_map<std::string, std::string> dna_map;
};

int main(int argc, const char* argv[]) {
    // Increase speed of input/output to standard streams
    std::ios::sync_with_stdio(false);
    std::cin.tie(0);

    // Initialize options to defaults
	std::string dna_filename;
	
    // Parse command line
    util::options::Parser parser("< prot_fasta > dna_fasta",
								 "Convert a protein alignment to its "
								 "equivalent dna alignment");
	parser.addStoreArg("dna_filename",
					   "name of fasta-formatted file containing DNA sequences",
					   dna_filename);
    parser.parse(argv, argv + argc);
	
    try {
		fasta::InputStream in_stream(std::cin);
		fasta::OutputStream out_stream(std::cout);
		fasta::InputStreamIterator in_iter_begin(in_stream), in_iter_end;
		fasta::OutputStreamIterator out_iter(out_stream);
		std::transform(in_iter_begin, in_iter_end, out_iter,
					   AlignmentConverter(dna_filename));
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

    return EXIT_SUCCESS;
}
