/* Copyright (c) 2007
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <vector>
#include <algorithm>

#include "boost/algorithm/string/predicate.hpp"

#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/util/options.hh"
using namespace deweylab;
using namespace bio::formats::fasta;

template<class InputIterator, class Compare>
bool is_sorted(InputIterator first, InputIterator last, Compare comparer) {
	if (first == last) { return true; }
	InputIterator prev(first);
	++first;
	while (first != last) {
		if (comparer(*first, *prev)) {
			return false;
		}
		prev = first;
		++first;
	}

	return true;
}

class FastaRecordComparer {
public:
	enum SortKey {TITLE, SEQUENCE, LENGTH};

	FastaRecordComparer(SortKey key, bool ignore_case)
		: key(key), ignore_case(ignore_case) {
	}
	
protected:
	SortKey key;
	bool ignore_case;

	bool is_less(const std::string& s1, const std::string& s2) const {
		return (ignore_case ?
				boost::ilexicographical_compare(s1, s2) :
				boost::lexicographical_compare(s1, s2));
	}

	bool is_equal(const std::string& s1, const std::string& s2) const {
		return (ignore_case ? boost::iequals(s1, s2) : boost::equals(s1, s2));
	}
	
	bool is_less(const Record& r1, const Record& r2) const {
		if (key == TITLE or key == SEQUENCE) {
			return is_less(key == TITLE ? r1.id : r1.sequence,
						   key == TITLE ? r2.id : r2.sequence);
		} else { // key == LENGTH
			return r1.sequence.length() < r2.sequence.length();
		}
	}

	bool is_equal(const Record& r1, const Record& r2) const {
		if (key == TITLE or key == SEQUENCE) {
			return is_equal(key == TITLE ? r1.id : r1.sequence,
							key == TITLE ? r2.id : r2.sequence);
		} else { // key == LENGTH
			return r1.sequence.length() == r2.sequence.length();
		}
	}
};

class FastaRecordLessThan : public FastaRecordComparer {
public:
	FastaRecordLessThan(SortKey key, bool ignore_case, bool reverse) 
		: FastaRecordComparer(key, ignore_case), reverse(reverse) {
	}
	
	bool operator()(const Record& r1, const Record& r2) const {
		return reverse ? is_less(r2, r1) : is_less(r1, r2);
	}

private:
	bool reverse;
};

class FastaRecordEquals : public FastaRecordComparer {
public:
	FastaRecordEquals(SortKey key, bool ignore_case)
		: FastaRecordComparer(key, ignore_case) {
	}
	
	bool operator()(const Record& r1, const Record& r2) const {
		return is_equal(r1, r2);
	}
};

int main(int argc, const char* argv[]) {
	// Increase speed of input/output to standard streams
	std::ios::sync_with_stdio(false);

	// Initialize options to defaults
	bool ignore_case = false;
	FastaRecordComparer::SortKey key = FastaRecordComparer::TITLE;
	bool reverse = false;
	bool unique = false;
	bool check = false;

	// Parse command line
	util::options::Parser parser("< fastaInput", "Sort FASTA records by title");
	parser.addStoreConstOpt('t', "title", "Sort by title (default)",
							key, FastaRecordComparer::TITLE);
	parser.addStoreConstOpt('s', "sequence", "Sort by sequence",
							key, FastaRecordComparer::SEQUENCE);
	parser.addStoreConstOpt('l', "length", "Sort by sequence length",
							key, FastaRecordComparer::LENGTH);
	parser.addStoreTrueOpt('r', "reverse", "Reverse sort order",
						   reverse);
	parser.addStoreTrueOpt('i', "ignore-case", "Ignore case",
						   ignore_case);
	parser.addStoreTrueOpt('u', "unique",
						   "Output the first of any set of records that "
						   "compare equal according to the sort key",
						   unique);
	parser.addStoreTrueOpt('c', "check", "Just check if the input is sorted",
						   check);
	parser.parse(argv, argv + argc);

	try {
		// Construct FASTA streams and iterators for fast reading and writing
		InputStream in_stream(std::cin);
		InputStreamIterator in_begin(in_stream), in_end;

		FastaRecordLessThan rec_less_than(key, ignore_case, reverse);
		
		if (check) {
			return (::is_sorted(in_begin, in_end, rec_less_than) ?
					EXIT_SUCCESS : EXIT_FAILURE);
		} else {
			OutputStream out_stream(std::cout);
			OutputStreamIterator out_begin(out_stream);
			std::vector<Record> recs;
		
			// Read all sequences into vector
			std::copy(in_begin, in_end, std::back_inserter(recs));

			// Sort sequences by specified key
			std::sort(recs.begin(), recs.end(), rec_less_than);

			// Output sorted sequences
			if (unique) {
				FastaRecordEquals rec_equals(key, ignore_case);
				std::unique_copy(recs.begin(), recs.end(), out_begin,
								 rec_equals);
			} else {
				std::copy(recs.begin(), recs.end(), out_begin);
			}
		}
			
	} catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}
		
	return EXIT_SUCCESS;
}
