/* Copyright (c) 2010
   Colin Dewey (University of Wisconsin-Madison)
   cdewey@biostat.wisc.edu
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "deweylab/util/options.hh"
#include "deweylab/filesystem.hh"

#include "deweylab/bio/formats/fasta.hh"
#include "deweylab/bio/formats/qual.hh"
#include "deweylab/bio/formats/fastq.hh"

#include "boost/shared_ptr.hpp"

#include <vector>
#include <string>

using namespace deweylab;
using namespace deweylab::bio::formats;
using fastq::PhredScore;
using fastq::QualityScoreEncoding;
using boost::shared_ptr;
using filesystem::OutputFileStream;

std::vector<PhredScore> fastq_phred_scores(const fastq::Record& fastq_rec) {
    std::vector<PhredScore> scores(fastq_rec.qualities.size());
    for (size_t i = 0; i < scores.size(); ++i) {
        scores[i] = fastq_rec.phred_quality(i);
    }
    return scores;
}

qual::Record qual_from_fastq(const fastq::Record& fastq_rec) {
    return qual::Record(fastq_rec.id,
                        fastq_rec.description,
                        fastq_phred_scores(fastq_rec));
}

fasta::Record fasta_from_fastq(const fastq::Record& fastq_rec) {
    return fasta::Record(fastq_rec.id,
                         fastq_rec.description,
                         fastq_rec.sequence);
}
    
int main(int argc, const char* argv[]) {
    std::string qual_filename;
    std::string encoding_name = "sanger";

    util::options::Parser parser;
    parser.addStoreOpt('q', "quals-file",
                       "File to which to output quality scores in QUAL format",
                       qual_filename);
    parser.addStoreOpt('e', "encoding",
                       "Quality score encoding scheme",
                       encoding_name);
	parser.parse(argv, argv + argc);
    
	try {
        const QualityScoreEncoding* encoding =
            fastq::get_quality_score_encoding(encoding_name);
        fastq::InputStream fastq_in(std::cin, encoding);
        fasta::OutputStream fasta_out(std::cout);

        shared_ptr<filesystem::OutputFileStream> qual_file;
        shared_ptr<qual::OutputStream> qual_out;
        if (!qual_filename.empty()) {
            qual_file = shared_ptr<OutputFileStream>(new OutputFileStream(qual_filename));
            qual_out = shared_ptr<qual::OutputStream>(new qual::OutputStream(*qual_file));
        }
        
        fastq::Record fastq_rec;
        while (fastq_in >> fastq_rec) {
            fasta_out << fasta_from_fastq(fastq_rec);
            if (qual_out != NULL) {
                *qual_out << qual_from_fastq(fastq_rec);
            }
        }
    } catch (const std::runtime_error& e) {
		std::cerr << "Error: " << e.what() << '\n';
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
